#include "userInterface.hpp"

#include <stdlib.h>                 // atoi()
#include <stdio.h>                  // printf() or iprintf()
#include <string.h>                 // strtok()

#include "chip_select_code.h"

#include "osHandles.h"
#include "utils/rprintf.h"     // rprintf
#include "fat/ff.h"            // FAT File System Library
#include "fat/diskio.h"        // Disk IO Initialize SPI Mutex

#include "drivers/uart0.h"     // uart0GetChar()
#include "drivers/ssp_spi.h"
#include "tasks/shift_brite.h"

#define MATCH(a,b) (0 == strcmp(a,b))
void getUartLine(char* uartInput);
void printLine()
{
	rprintf("------------------------------------------------------------------------------------\n");
}

void writeEnable()
{
	CS_Sel(sel_Flash);
	char opCode = 0x06; // opCode for write Enable
	rxTxByteSSPSPI(opCode);
	CS_Sel(sel_Nothing); // de-select flash
}
void globalProtectionDisable()
{
	CS_Sel(sel_Flash);
	char opCode = 0x01; // opCode for write status register
	rxTxByteSSPSPI(opCode);
	rxTxByteSSPSPI(0x00); //// send one byte dummy
	CS_Sel(sel_Nothing);
}
void waitUntilNotBusy()
{
	CS_Sel(sel_Flash);
	char opCode = 0x05;
	rxTxByteSSPSPI(opCode);
	while(char mID=rxTxByteSSPSPI(0x00))
	{
		//rprintf("%X \n", mID);
		if(mID==0x11) // if 1 means is not ready
			continue;
		else
			break;
	}
	CS_Sel(sel_Nothing); // de-select flash
}
void eraseBlock(int address)
{
	typedef union
	{
		int raw;
		struct{
				char byte0;
				char byte1;
				char byte2;
				char byte3;
			   };
	}IntSplitter;
	IntSplitter splitter;
	splitter.raw = address ;
	CS_Sel(sel_Flash); // select flash
	char opCode = 0x20;
	rxTxByteSSPSPI(opCode);
	//rxTxByteSSPSPI(opCode);
	rxTxByteSSPSPI(splitter.byte2);
	rxTxByteSSPSPI(splitter.byte1);
	rxTxByteSSPSPI(splitter.byte0);
	CS_Sel(sel_Nothing); // de-select flash
}
void writePage(int address,char *data,int length)
{
	typedef union
	{
		int raw;
		struct{
				char byte0;
				char byte1;
				char byte2;
				char byte3;
			   };
	}IntSplitter;
	IntSplitter splitter;
	splitter.raw = address ;
	//rprintf("\n");
	//rprintf("splitter byte2 is %X\n",splitter.byte2);
	//rprintf("splitter byte1 is %X\n",splitter.byte1);
	//rprintf("splitter byte0 is %X\n",splitter.byte0);
	CS_Sel(sel_Flash);
	char opCode = 0x02;
	rxTxByteSSPSPI(opCode);

	rxTxByteSSPSPI(splitter.byte2);
	rxTxByteSSPSPI(splitter.byte1);
	rxTxByteSSPSPI(splitter.byte0);

	while(length--)
	{
		rxTxByteSSPSPI(*data++);
		//rprintf("%c", data[i]); //debugging purposes.

	}
	rxTxByteSSPSPI(0);
	CS_Sel(sel_Nothing);
	rprintf("\n");
}
void readPage(char *buff, int address)
{
	typedef union
	{
		int raw;
		struct{
				char byte0;
				char byte1;
				char byte2;
				char byte3;
				};
	}IntSplitter;
	IntSplitter splitter;
	splitter.raw = address;
	CS_Sel(sel_Flash); // select flash
	char opCode = 0x0B;
	//rprintf("\n");
	//rprintf("splitter byte2 is %X\n",splitter.byte2);
	//rprintf("splitter byte1 is %X\n",splitter.byte1);
	//rprintf("splitter byte0 is %X\n",splitter.byte0);
	rxTxByteSSPSPI(opCode);
	rxTxByteSSPSPI(splitter.byte2);
	rxTxByteSSPSPI(splitter.byte1);
	rxTxByteSSPSPI(splitter.byte0);
	rxTxByteSSPSPI(0x00);
	rprintf("\n");
	 do{
	         *buff = rxTxByteSSPSPI(0x00);
	          rprintf("%c", *buff);
	    }while(*buff++!=0);

	 rprintf("\n");
	CS_Sel(sel_Nothing);
}
void shift_brite_init()
{
    PINSEL0 &= ~(0xFF<<8);
    //P0.4 to P0.7 to OUTPUT
    IODIR0 |= (1<<4);
    IODIR0 |= (1<<5);
    IODIR0 |= (1<<6);
    IODIR0 |= (1<<7);
    //set the enable output (P0.6) low to enable the shiftbrite
    IOCLR0 = (1<<6);
}

void uartUI(void *pvParameters)
{
	OSHANDLES *osHandles = (OSHANDLES*)(pvParameters);
    char uartInput[128];
    CS_Init(); // initiaize chip select
    initialize_SSPSPI();
    diskio_initializeSPIMutex(&(osHandles->lock.SPI));
    initialize_SdCardSignals();
    shift_brite_init();
	//IOCLR0 = (1<<5);
	rprintf("shiftbrite is initialized\n");

	FATFS SDCard; // Takes ~550 Bytes
	if(FR_OK != f_mount(0, &SDCard)) { // Mount the Card on the File System
		rprintf("WARNING: SD CARD Could not be mounted!\n");
	}

	for (;;)
	{
		rprintf("LPC: ");
		memset(uartInput, 0, sizeof(uartInput));
		getUartLine(uartInput);
		if (strlen(uartInput) <= 0){
			continue;
		}

		char *command = strtok(uartInput, " ");

		if (MATCH(command, "echo")){
			char *echoBack = strtok(NULL, "");
			rprintf("Echo: %s\n", echoBack);
		}
		else if(MATCH(command, "set"))
		{
			char * color = strtok(0, "");
			if (MATCH(color,"red"))
			{
				//rprintf("testing\n");
				latchHigh();
				sendPacket(colorPacket(600,0,0));
				vTaskDelay(2000);
				latchLow();
			}
			else if (MATCH(color,"off"))
			{
				//rprintf("testing\n");
				latchHigh();
				sendPacket(colorPacket(0,0,0));
				vTaskDelay(2000);
				latchLow();
			}
			else if (MATCH(color,"green"))
			{
				//rprintf("testing\n");
				latchHigh();
				sendPacket(colorPacket(0,1000,0));
				vTaskDelay(2000);
				latchLow();
			}
			else if (MATCH(color,"blue"))
			{
				//rprintf("testing\n");
				latchHigh();
				sendPacket(colorPacket(0,0,1000));
				vTaskDelay(2000);
				latchLow();
			}
			else if (MATCH(color, "yellow"))
			{
				latchHigh();
				sendPacket(colorPacket(1000,825,0));
				vTaskDelay(2000);
				latchLow();
			}
			else if (MATCH(color,"changing"))
			{
				latchHigh();
				  sendPacket(colorPacket(1023,0,0));
				  vTaskDelay(1000);
				  latchLow();
				  sendPacket(colorPacket(0,1023,0));
				  vTaskDelay(1000);
				  latchLow();
				  sendPacket(colorPacket(0,0,1023));
				  vTaskDelay(1000);
				  //vTaskDelay(2000);
				  latchHigh();
				  latchLow();
			}
		}
		// TODO 5.  Add your Flash Read/Write Functions here
		else if(MATCH(command, "WriteFlash"))
		{
			int loc = atoi(strtok (0, " ")) * 256;
			char * data = strtok(0, "");
			int dataSize = strlen(data);
			//rprintf("%i %s %i \n", loc, data, dataSize);

			if(xSemaphoreTake(osHandles->lock.SPI,3000)) // take semaphore
			{
				CS_Sel(sel_Flash); // select flash
				writeEnable();
				globalProtectionDisable();
				waitUntilNotBusy();
				writeEnable();
				eraseBlock(loc);
				waitUntilNotBusy();
				writeEnable();
				writePage(loc, data, dataSize); // pass the values into the writePage function

				CS_Sel(sel_Nothing); // de-select flash
				xSemaphoreGive(osHandles->lock.SPI);
			}
			else
			{
				rprintf("Erase Error: SPI semaphore \n");
			}
		}
		else if(MATCH(command,"ReadFlash"))
		{
			char buff[256];
			int loc = atoi(strtok (0, " ")) * 256;
			if(xSemaphoreTake(osHandles->lock.SPI,3000)) // take semaphore
			{
				readPage(buff,loc);
				xSemaphoreGive(osHandles->lock.SPI); //give semaphore back if time expires
			}
			else
			{
				rprintf("ReadFlash Error: SPI semaphore \n");
			}
		}
		else if(MATCH(command,"ReadSig"))
		{
			//int loc = atoi(strtok (0, ""));
			//rprintf("Location was %i \n", loc);
			if(xSemaphoreTake(osHandles->lock.SPI,3000)) // take semaphore
			{
				CS_Sel(sel_Flash); // select flash
				char opCode = 0x9F;
				rxTxByteSSPSPI(opCode);
				char mID = rxTxByteSSPSPI(0x00); // send garbage byte
				char dID1 = rxTxByteSSPSPI(0x01); // send garbage byte
				char dID2 = rxTxByteSSPSPI(0x02); // send garbage byte
				char eID = rxTxByteSSPSPI(0x03); // send garbage byte
				rprintf("\nID Info: %X, %X, %X, %X \n", mID,dID1,dID2,eID);
				CS_Sel(sel_Nothing); //de-select flash
				xSemaphoreGive(osHandles->lock.SPI); //give semaphore back if time expires
			}
			else
			{
				rprintf("ReadSig Error: SPI semaphore \n");
			}
		}
		else if(MATCH("ls", command))
		{
			DIR Dir;
			FILINFO Finfo;
			FATFS *fs;
			FRESULT returnCode = FR_OK;

			unsigned int fileBytesTotal, numFiles, numDirs;
			fileBytesTotal = numFiles = numDirs = 0;
			#if _USE_LFN
			char Lfname[512];
			#endif

			char dirPath[] = "0:";
			if (FR_OK != (returnCode = f_opendir(&Dir, dirPath))) {
				rprintf("Invalid directory: |%s|\n", dirPath);
				continue;
			}

			rprintf("Directory listing of: %s\n\n", dirPath);
			for (;;) {
				#if _USE_LFN
				Finfo.lfname = Lfname;
				Finfo.lfsize = sizeof(Lfname);
				#endif

				char returnCode = f_readdir(&Dir, &Finfo);
				if ( (FR_OK != returnCode) || !Finfo.fname[0])
					break;
				if (Finfo.fattrib & AM_DIR){
					numDirs++;
				}
				else{
					numFiles++;
					fileBytesTotal += Finfo.fsize;
				}
				iprintf("%c%c%c%c%c %u/%2u/%2u %2u:%2u %10lu %13s",
						(Finfo.fattrib & AM_DIR) ? 'D' : '-',
						(Finfo.fattrib & AM_RDO) ? 'R' : '-',
						(Finfo.fattrib & AM_HID) ? 'H' : '-',
						(Finfo.fattrib & AM_SYS) ? 'S' : '-',
						(Finfo.fattrib & AM_ARC) ? 'A' : '-',
						(Finfo.fdate >> 9) + 1980, (Finfo.fdate >> 5) & 15, Finfo.fdate & 31,
						(Finfo.ftime >> 11), (Finfo.ftime >> 5) & 63, Finfo.fsize, &(Finfo.fname[0]));
				#if _USE_LFN
				iprintf(" -- %s", Lfname);
				#endif
				iprintf("\n");
			}
			iprintf("\n%4u File(s), %10u bytes total\n%4u Dir(s)", numFiles, fileBytesTotal, numDirs);

			if (f_getfree(uartInput, (DWORD*) &fileBytesTotal, &fs) == FR_OK)
			{
				iprintf(", %10uK bytes free\n", fileBytesTotal * fs->csize / 2);
			}
		}
		else if(MATCH(command, "read")) {
			char *filename = strtok(0, "");
			rprintf("File to read: |%s|\n", filename);

			FIL file;
			if(FR_OK == f_open(&file, filename, (FA_READ | FA_OPEN_EXISTING)))
			{
				char buff[513] = {0};
				unsigned int bytesRead = 0;
				rprintf("File Contents:\n");
				printLine();
				do
				{
					if(FR_OK != f_read(&file, buff, sizeof(buff)-1, &bytesRead))
					{
						rprintf("--Failed to read file after having it opened--\n");
						break;
					}
					rprintf("%s", buff);
				}while(0); // }while(bytesRead == bytesToRead); <- Don't want to read entire file in a loop
				printLine();

				f_close(&file);
				rprintf("File is closed\n");
			}
			else
			{
				rprintf("Failed to open the file\n");
			}
		}
		else if(MATCH(command, "write")) {
			char *filename = strtok(0, "");
			rprintf("File to write: |%s|\n", filename);

			FIL file;
			if(FR_OK == f_open(&file, filename, (FA_WRITE | FA_CREATE_ALWAYS)))
			{
				rprintf("File successfully opened\n");
				char buff[] = "File written using FATFS From Chen\n";

				unsigned int bytesWritten = 0;
				if(FR_OK != f_write(&file, buff, sizeof(buff), &bytesWritten))
				{
					rprintf("Failed to write file\n");
				}

				f_close(&file);
				rprintf("File is closed\n");
			}
			else
			{
				rprintf("Failed to open the file\n");
			}
		}
#if configGENERATE_RUN_TIME_STATS
		else if(MATCH(command, "CPU")) {
			char buff[512];
			vTaskGetRunTimeStats(buff, 0);
			rprintf("%s", buff);
		}
		else if(MATCH(command, "CPUr")) {
			char buff[512];
			vTaskGetRunTimeStats(buff, 1);
			rprintf("%s", buff);
		}
		else if(MATCH(command, "CPUn")) {
			int delayTimeMs = atoi(strtok(NULL, ""));
			if(delayTimeMs > 0) {
				char buff[512];
				rprintf("Delaying for %ims.  CPU Usage will be reported afterwards...\n", delayTimeMs);
				vTaskGetRunTimeStats(buff, 1);
				vTaskDelay(OS_MS(delayTimeMs));
				vTaskGetRunTimeStats(buff, 1);
				rprintf("%s", buff);
			}
			else {
				rprintf("You must define a delay time in milliseconds as parameter.\n");
			}
		}
#endif
#if configUSE_TRACE_FACILITY != 0
		else if(MATCH(command, "info")) {
			char buff[512];
			vTaskList(buff);
			rprintf("%s", buff);
		}
#endif
		else if (MATCH("HELP", command)) {
			rprintf("\n");
			printLine();
			rprintf("Command list:\n");
			printLine();
			rprintf("echo <anything>    - Echoes back anything sent.\n");
			rprintf("ls                 - Get directory listing.\n");
			rprintf("read <filename>    - Read a file.\n");
			rprintf("write <filename>   - Wrie a file.\n");
				#if configGENERATE_RUN_TIME_STATS
			rprintf("CPU                - Show tasks and CPU Allocation.\n");
			rprintf("CPUr               - Show tasks and CPU Allocation & reset counters\n");
			rprintf("CPUn <# ms>        - Reset counters, delay, then report CPU usage & reset counters\n");
				#endif
				#if configUSE_TRACE_FACILITY != 0
			rprintf("info               - Show tasks and their Stack-high watermark.\n");
				#endif
			printLine();
		}
		else {
			rprintf("Command not recognized.\n");
		}
	}
}

void getUartLine(char* uartInput)
{
	char data;
	unsigned int uartInputPtr = 0;

	while (1)
	{
		while(!uart0GetChar(&data, portMAX_DELAY));

		if (data == '\n'){
			uartInput[uartInputPtr] = 0;
			uart0PutChar('\n', 100);
			break;
		}
		else if (data == '\r'){
			// Ignore it
		}
		else if (data == '\b')
		{
			if (uartInputPtr > 0){
				uartInputPtr--;
				rprintf("\b \b");
			}
			else {					// Nothing to backspace
				uart0PutChar(7, 0);	// ASCII Char 7 is for bell
			}
		}

		else{
			uart0PutChar(data, 100);
			uartInput[uartInputPtr++] = data;
		}
	}
}
