#include <string.h>					// memset, memcpy
#include "../System/lpc214x.h"
#include "../System/sysConfig.h"	// Need PCLK to setup I2C Rate.
#include "../FreeRTOS/FreeRTOS.h"
#include "../FreeRTOS/Semphr.h"
#include "../FreeRTOS/task.h"     // runTimeStats() functions
#include "../drivers/i2c.h"
#include <algorithm>
#include "adjd_s311.hpp"
#include "../utils/rprintf.h"

using namespace std;

RGBC calibrate()
{
	//delay(2); //wait for LED to come full brintness if used
	RGBC color;
	rprintf("half value COLOR is %i\n", calibrateColor()); // This calibrates R, G, and B int registers
	rprintf("half value CLEAR is %i\n", calibrateClear()); // This calibrates the C int registers
	//rprintf("diff CAP is %i\n", calibrateCapacitors()); // This calibrates the RGB, and C cap registers

	return color;
}

void init_colorsensor()
{
	unsigned char colorCap[4];
	unsigned int colorInt[4];
	colorCap[RED] = 15;
	colorCap[GREEN] = 15;
	colorCap[BLUE] = 15;
	colorCap[CLEAR] = 15;
	// values must be between 0 and 15

	colorInt[RED] = 1000;
	colorInt[BLUE] = 1000;
	colorInt[GREEN] = 1000;
	colorInt[CLEAR] = 10;
	// max value for these is 4095

	// sensor gain registers, CAP_... to select number of capacitors.
	// value must be <= 15 */
	writeRegister(colorCap[RED] & 0xF, CAP_RED);
	writeRegister(colorCap[GREEN] & 0xF, CAP_GREEN);
	writeRegister(colorCap[BLUE] & 0xF, CAP_BLUE);
	writeRegister(colorCap[CLEAR] & 0xF, CAP_CLEAR);

	//Write sensor gain registers INT_...
	//to select integration time  value must be <= 4096
	writeRegister((unsigned char) colorInt[RED], INT_RED_LO);
	writeRegister((unsigned char) (colorInt[RED] >> 8), INT_RED_HI);
	writeRegister((unsigned char) colorInt[BLUE], INT_BLUE_LO);
	writeRegister((unsigned char) (colorInt[BLUE] >> 8), INT_BLUE_HI);
	writeRegister((unsigned char) colorInt[GREEN], INT_GREEN_LO);
	writeRegister((unsigned char) (colorInt[GREEN] >> 8), INT_GREEN_HI);
	writeRegister((unsigned char) colorInt[CLEAR], INT_CLEAR_LO);
	writeRegister((unsigned char) (colorInt[CLEAR] >> 8), INT_CLEAR_HI);
}

/* calibrateClear() - This function calibrates the clear integration registers
 of the ADJD-S311.
 */
int calibrateClear()
{
	int gainFound = 0;
	int upperBox = 4096;
	int lowerBox = 0;
	int half;

	while (!gainFound)
	{
		half = ((upperBox - lowerBox) / 2) + lowerBox;
		//no further halfing possbile

		if (half == lowerBox)
			gainFound = 1;
		else
		{
			writeInt(INT_CLEAR_LO, half);
			performMeasurement();
			int halfValue = readRegisterInt(DATA_CLEAR_LO);

			if (halfValue > 1000)
				upperBox = half;
			else if (halfValue < 1000)
				lowerBox = half;
			else
				gainFound = 1;
		}
	}
	return half;
}

/* calibrateColor() - This function calibrates the RG and B
 integration registers.
 */
int calibrateColor()
{
	int gainFound = 0;
	int upperBox = 4096;
	int lowerBox = 0;
	int half = 4096;
	int red, green, blue;

	while (!gainFound)
	{
		half = ((upperBox - lowerBox) / 2) + lowerBox;
		//no further halfing possbile
		if (half == lowerBox)
			gainFound = 1;
		if(half == 0)
			gainFound = 1;
		else
		{
			writeInt(INT_RED_LO, half);
			writeInt(INT_GREEN_LO, half);
			writeInt(INT_BLUE_LO, half);

			performMeasurement();
			int halfValue = 0;
			red = readRegisterInt(DATA_RED_LO);
			green = readRegisterInt(DATA_GREEN_LO);
			blue = readRegisterInt(DATA_BLUE_LO);

			rprintf("red = %i, blue = %i, green = %i\n", red, blue, green);
			rprintf("half value = %i\n", half);
			halfValue = max(halfValue, red);
			halfValue = max(halfValue, green);
			halfValue = max(halfValue, blue);
			if (halfValue > 30000)
				//half = half - 1;
				upperBox = half;
			else if (halfValue < 30000)
				lowerBox = half;
			else
				gainFound = 1;

		}
	}
	rprintf("red = %i, blue = %i, green = %i\n", red, blue, green);
				rprintf("half value = %i\n", half);
		writeInt(INT_RED_LO, half);
		writeInt(INT_GREEN_LO, half);
		writeInt(INT_BLUE_LO, half);
	return half;
}

/* calibrateCapacitors() - This function calibrates each of the RGB and C
 capacitor registers.
 */
int calibrateCapacitors()
{
	int calibrationRed = 0;
	int calibrationBlue = 0;
	int calibrationGreen = 0;
	int calibrated = 0;

	//need to store detect better calibration
	int oldDiff = 5000;

	while (!calibrated)
	{
		// sensor gain setting (Avago app note 5330)
		// CAPs are 4bit (higher value will result in lower output)
		writeRegister(calibrationRed, CAP_RED);
		writeRegister(calibrationGreen, CAP_GREEN);
		writeRegister(calibrationBlue, CAP_BLUE);

		int colorGain = readRegisterInt(INT_RED_LO);
		writeInt(INT_RED_LO, colorGain);
		writeInt(INT_GREEN_LO, colorGain);
		writeInt(INT_BLUE_LO, colorGain);

		int maxRead = 0;
		int minRead = 4096;
		int red = 0;
		int green = 0;
		int blue = 0;

		for (int i = 0; i < 4; i++)
		{
			performMeasurement();
			red += readRegisterInt(DATA_RED_LO);
			green += readRegisterInt(DATA_GREEN_LO);
			blue += readRegisterInt(DATA_BLUE_LO);
		}
		red /= 4;
		green /= 4;
		blue /= 4;

		maxRead = max(maxRead, red);
		maxRead = max(maxRead, green);
		maxRead = max(maxRead, blue);

		minRead = min(minRead, red);
		minRead = min(minRead, green);
		minRead = min(minRead, blue);

		int diff = maxRead - minRead;

		if (oldDiff != diff)
		{
			if ((maxRead == red) && (calibrationRed < 15))
				calibrationRed++;
			else if ((maxRead == green) && (calibrationGreen < 15))
				calibrationGreen++;
			else if ((maxRead == blue) && (calibrationBlue < 15))
				calibrationBlue++;
			else if(calibrationBlue == 15)
				calibrated = 1;
		}

		else
			calibrated = 1;

		oldDiff = diff;

		int rCal = calibrationRed;
		int gCal = calibrationGreen;
		int bCal = calibrationBlue;
		rprintf("red cap = %i, blue cap = %i, green cap = %i\n", rCal, gCal, bCal);
	}
	return oldDiff;

}

/* writeInt() - This function writes a 12-bit value
 to the LO and HI integration registers */
void writeInt(int address, int gain)
{
	if (gain < 4096)
	{
		char msb = gain >> 8;
		char lsb = gain;

		writeRegister(lsb, address);
		writeRegister(msb, address + 1);
	}
}

/* performMeasurement() - This must be called before
 reading any of the data registers. This commands the
 ADJD-S311 to perform a measurement, and store the data
 into the data registers.*/
void performMeasurement()
{
	writeRegister(0x01, 0x00); // start sensing by sending 0x01 to the CTRL register
	while (readRegister(0x00) != 0)
		rprintf("not ready\n"); // waiting for a result
}

RGBC read()
{
	//Read data registers and return a RGBC var
	RGBC color = RGBC();

	performMeasurement();

	color.red = readRegister(DATA_RED_LO);
	color.green = readRegister(DATA_GREEN_LO);
	color.blue = readRegister(DATA_BLUE_LO);
	color.clear = readRegister(DATA_CLEAR_LO);

	return color;
}

/* getOffset() - This function performs the offset reading
 and stores the offset data into the colorOffset[] array.
 You can turn on data trimming by uncommenting out the
 writing 0x01 to 0x01 code.
 */
void getOffset()
{
	signed char colorOffset[4];
	//digitalWrite(_led_pin, LOW);  // turn LED off
	//delay(10);  // wait a tic
	writeRegister(0x02, 0x00); // start sensing
	while (readRegister(0x00) != 0)
		; // waiting for a result
	//writeRegister(0x01, 0x01);  // set trim
	//delay(100);
	for (int i = 0; i < 4; i++)
		colorOffset[i] = (signed char) readRegister(OFFSET_RED + i);
	//digitalWrite(_led_pin, HIGH);
}

// Write a byte of data to a specific ADJD-S311 address
void writeRegister(unsigned char data, unsigned char address)
{
	i2cWriteDeviceRegister(ADJD_S311_ADDRESS, address, data);
}

// read a byte of data from ADJD-S311 address
unsigned char readRegister(unsigned char address)
{
	i2cWriteDeviceRegister(ADJD_S311_ADDRESS, address, ADJD_S311_ADDRESS | (1
			<< 0));
	return i2cReadDeviceRegister(ADJD_S311_ADDRESS | (1 << 0), address);
}

int readRegisterInt(unsigned char address)
{
	i2cWriteDeviceRegister(ADJD_S311_ADDRESS, address, ADJD_S311_ADDRESS | (1<< 0));
	int lowerByte = i2cReadDeviceRegister(ADJD_S311_ADDRESS | (1 << 0), address);

	i2cWriteDeviceRegister(ADJD_S311_ADDRESS, address + 1, ADJD_S311_ADDRESS | (1<< 0));
	int upperByte = i2cReadDeviceRegister(ADJD_S311_ADDRESS | (1 << 0), address + 1);

	int value = lowerByte | (upperByte << 8);
	return value;
}

